#!/usr/bin/env python
#----------------------------------------------------------------------------
# Name:         SimApp.py
# Purpose:      A basic, somewhat generic simulator application
#
# Author:       Matthew Patitz
#
# Created:      
# RCS-ID:       
# Copyright:    
# License:      
#----------------------------------------------------------------------------

"""
This is a generic application for simulation programs

Requirements:
    This application requires Python 2.x

Dependencies:
    wxPython    (download installer for your platform here: https://wxpython.org/Phoenix/snapshot-builds/)
"""

#MJP TODO:
#    - probably no longer need to disable mouse-over during fast forward
#    - visual display of frontier & growth locations(?)
#    - implement save
#        - as txt
#        - as image
#            - show grid in saved image?
#    - grip scrolling
#
#Previous dependencies no longer used:
#    pyPdf2      (install with 'pip install pypdf2')
#    Pillow      (install with 'pip install Pillow')
#    reportlab   (install with 'pip install reportlab')


#import wxversion
#wxversion.select('3.0-msw', optionsRequired=True)
#wxversion.select('3.0-Phoenix', optionsRequired=True)
import wx
import wx.richtext as rt
import os
import random
import ast
import sys
import thread
import threading
import wx.aui
from wx.lib.pubsub import pub
#from wx.lib.pubsub import Publisher
import SimUtils
import SimCanvas


import wx.richtext as rtcm 

class SimRTC(rtcm.RichTextCtrl): 

    def __init__(self, parent, app, id=-1, style=wx.VSCROLL|wx.HSCROLL|rtcm.RE_READONLY, pos=wx.DefaultPosition, size=wx.DefaultSize): 
        rtcm.RichTextCtrl.__init__(self, parent, id=-1, style=style, pos=pos, size=size) 
        self.Bind(wx.EVT_CHAR, app.canvas.OnChar)
        self.Bind(wx.EVT_MOUSE_EVENTS, self.MouseHandler)

    def MouseHandler(self, evt):
        # Comment out the following line to disable clicking into this window, however note that this disallows selection and copying of text
        evt.Skip()
        return

    def VisibleWrite(self, msg):  
        self.WriteText(msg)
        if "wxMSW" in wx.PlatformInfo:
            self.ScrollLines(self.GetNumberOfLines()*10)
        else:   # Horrible hack to get it to scroll to the bottom on OS X too
            sp = self.messageWin.GetScrollPos(wx.VERTICAL)
            #r = self.messageWin.GetScrollRange(wx.VERTICAL) - self.messageWin.GetScrollThumb(wx.VERTICAL)
            e = wx.ScrollEvent(wx.wxEVT_SCROLLWIN_LINEDOWN,
                              self.messageWin.GetId(),
                              10,
                              wx.VERTICAL)
            for i in range(10):
                self.messageWin.GetEventHandler().ProcessEvent(e)
    

#assertMode = wx.PYAPP_ASSERT_DIALOG

# IDs
SIMFW_ID_STOP          = 100
SIMFW_ID_FOR           = 101
SIMFW_ID_FF            = 102
SIMFW_ID_ZOOMIN        = 103
SIMFW_ID_ZOOMOUT       = 104
SIMFW_ID_ZOOM_MODE     = 105
SIMFW_ID_SELECT_MODE   = 106
SIMFW_ID_DELETE        = 107
SIMFW_ID_UNDO          = 108



#---------------------------------------------------------------------------


class SimApp(wx.App):
    def __init__(self):
        SimUtils.__init__(self)
        wx.App.__init__(self, redirect=False)

    def OnInit(self, title="Simple self-assembly simulator application"):
        self.ctrlPanel = None
        self.messageWin = None
        self.showMessages = True
        self.messageQueue = []

        #wx.Log_SetActiveTarget(wx.LogStderr())
        #self.SetAssertMode(assertMode)

        self.frame = wx.Frame(None, -1, title, pos=(50,50), size=(400,400),
                         style=wx.DEFAULT_FRAME_STYLE, name=title)
        self.CreateStatusBar()
        self.frame._mgr = wx.aui.AuiManager(self.frame)

        self.CreateMenu()

        self.frame.Show(True)

        self.CreateCanvas()
        self.frame._mgr.AddPane(self.canvas, wx.CENTER, 'Simulation Window')

        self.AddMessageWindow()

        self.frame.Bind(wx.EVT_CLOSE, self.OnCloseFrame)
        self.frame.Bind(wx.EVT_CHAR, self.canvas.OnChar)
        
        self.CreateToolbar()
        
        pub.getDefaultPublisher().subscribe(self.DoOutputMessage, "outputMessage")
        pub.getDefaultPublisher().subscribe(self.DoWarningDialog, "warningDialog")
        pub.getDefaultPublisher().subscribe(self.DoErrorDialog, "errorDialog")

        # so set the frame to a good size for showing stuff
        self.frame.SetSize((1200, 800))
        self.frame.SetFocus()
        self.window = self.frame

        self.SetTopWindow(self.frame)

        #wx.Log_SetActiveTarget(wx.LogStderr())
        #wx.Log_SetTraceMask(wx.TraceMessages)
        self.frame.Maximize()
        
        # tell the manager to 'commit' all the changes just made
        self.frame._mgr.Update()
        
        #self.canvas.InitializeDrawing()
        #self.canvas.Refresh()
        self.canvas.DrawAssemblyUpdate("new")
        self.canvas.ScrollToCoords((0,0))

        return True
    
    def CreateStatusBar(self):
        self.frame.CreateStatusBar(3)
        self.frame.SetStatusWidths([150,200,-1])

    def CreateMenu(self):
        menuBar = wx.MenuBar()
        menu = wx.Menu()
        item1 = menu.Append(-1, "&Load\tCtrl-L", "Load system")
        self.Bind(wx.EVT_MENU, self.OnLoad, item1)
        itemR = menu.Append(-1, "Reload files", "Reload system from files")
        self.Bind(wx.EVT_MENU, self.OnReload, itemR)
        itemReset = menu.Append(-1, "&Reset to seed\tCtrl-R", "Reset system to seed")
        self.Bind(wx.EVT_MENU, self.OnReset, itemReset)
        item2 = menu.Append(-1, "&Save\tCtrl-S", "Save system")
        self.Bind(wx.EVT_MENU, self.OnSave, item2)
        menu.AppendSeparator()
        item3 = menu.Append(-1, "E&xit\tCtrl-X", "Exit simulator")
        self.Bind(wx.EVT_MENU, self.OnExitApp, item3)
        menuBar.Append(menu, "&File")

        menuCtrl = wx.Menu()
        itemStep = menuCtrl.Append(-1, "&Forward\tCtrl-F", "Forward one step")
        self.Bind(wx.EVT_MENU, self.OnForward, itemStep)
        itemFF = menuCtrl.Append(-1, "&Go\tCtrl-G", "Fast forward")
        self.Bind(wx.EVT_MENU, self.OnGo, itemFF)
        itemQuit = menuCtrl.Append(-1, "&Quit\tCtrl-Q", "Quit fast forward")
        self.Bind(wx.EVT_MENU, self.OnQuit, itemQuit)
        menuBar.Append(menuCtrl, "&Control")

        menuView = wx.Menu()
        itemSTC = menuView.Append(-1, "&Center\tCtrl+C", "Scroll to center")
        self.Bind(wx.EVT_MENU, self.OnCenter, itemSTC)
        menuView.AppendSeparator()
        itemZIn = menuView.Append(-1, "Zoom in\tCtrl-+", "Zoom in")
        self.Bind(wx.EVT_MENU, self.OnZoomIn, itemZIn)
        itemZOut = menuView.Append(-1, "Zoom out\tCtrl--", "Zoom out")
        self.Bind(wx.EVT_MENU, self.OnZoomOut, itemZOut)
        menuView.AppendSeparator()
        itemTG = menuView.Append(-1, "&Toggle grid lines\tCtrl-T", "Toggle display of grid lines")
        self.Bind(wx.EVT_MENU, self.OnToggleGrid, itemTG)
        itemTD = menuView.Append(-1, "&Toggle display updates\tCtrl-D", "Toggle display of tile additions")
        self.Bind(wx.EVT_MENU, self.OnToggleDisplay, itemTD)
        itemSTT = menuView.Append(-1, "Show only matching tile types...", "Set patterns for tile type names to display")
        self.Bind(wx.EVT_MENU, self.OnShowMatching, itemSTT)
        
        menuView.AppendSeparator()
        itemRefresh = menuView.Append(-1, "Refresh screen\tF2", "Refresh screen")
        self.Bind(wx.EVT_MENU, self.OnRefresh, itemRefresh)

        menuBar.Append(menuView, "&View")

        menu2 = wx.Menu()
        itemH = menu2.Append(-1, "&Help\tCtrl-H", "Show help")
        self.Bind(wx.EVT_MENU, self.OnHelp, itemH)
        menuBar.Append(menu2, "&Help")
        
        self.frame.SetMenuBar(menuBar)

    def AddMessageWindow(self):
        self.ctrlPanel = wx.Panel(self.frame, wx.ID_ANY, pos=wx.DefaultPosition, size=(300,200), style = wx.TAB_TRAVERSAL)
        self.messageWin = SimRTC(self.ctrlPanel, self, wx.ID_ANY, pos=wx.DefaultPosition, size=(500,200))#, style=wx.NOSCROLL|wx.HSCROLL|wx.NO_BORDER)
        #self.messageWin = rtcm.RichTextCtrl(self.ctrlPanel, id=wx.ID_ANY, style=wx.VSCROLL|wx.HSCROLL|rtcm.RE_READONLY, pos=wx.DefaultPosition, size=(500,70)) 
        #self.messageWin = rt.RichTextCtrl(self.ctrlPanel, wx.ID_ANY, pos=wx.DefaultPosition, size=(500,70))#, style=wx.NOSCROLL|wx.HSCROLL|wx.NO_BORDER)
        self.messageWin.SetEditable(False)
        sz = wx.BoxSizer(wx.HORIZONTAL)
        sz.Add(self.messageWin, 20, wx.EXPAND, 20)
        self.ctrlPanel.SetSizerAndFit(sz)
        self.frame._mgr.Bind(wx.aui.EVT_AUI_PANE_CLOSE, self.OnCloseMessagePane)
        self.frame._mgr.AddPane(self.ctrlPanel, wx.BOTTOM, 'Message Panel')
        self.frame._mgr.Update()

    def OnCloseMessagePane(self, event):
        if event.GetPane().window == self.ctrlPanel:
            self.menuView.Check(self.itemShowMess.GetId(), False)
        event.Skip()
        
    def OnShowMessageWin(self, event):
        pane = self.frame._mgr.GetPane(self.ctrlPanel)
        if self.itemShowMess.IsChecked():
            pane.Show()
        else:
            pane.Hide()
        self.frame._mgr.Update()

    def CreateCanvas(self):
        self.canvas = SimCanvas.Canvas(self.frame, wx.ID_ANY, self.frame, wx.DefaultSize, self)

    def CreateToolbar(self):
        self.toolbar = wx.ToolBar(self.frame, -1, wx.DefaultPosition, wx.DefaultSize, wx.TB_DOCKABLE | wx.TB_FLAT)
        self.toolbar.SetToolSeparation(20)
        self.toolbar.SetToolBitmapSize(wx.Size(24,24))
        self.toolbar.AddTool(SIMFW_ID_STOP, "Pause", wx.Image("images/pause.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Pause")#, longHelp="Pause simulation")
        self.toolbar.AddTool(SIMFW_ID_FOR, "Forward", wx.Image("images/right.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Forward")#, longHelp="Forward one step")
        self.toolbar.AddTool(SIMFW_ID_FF, "Fast Forward", wx.Image("images/fastright.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Fast Forward")#longHelp="Fast forward")
        self.toolbar.AddSeparator()
        self.toolbar.AddTool(SIMFW_ID_ZOOMIN, "Zoom In", wx.Image("images/zoomin.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Zoom In")#, longHelp="Zoom in")
        self.toolbar.AddTool(SIMFW_ID_ZOOMOUT, "Zoom Out", wx.Image("images/zoomout.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Zoom Out")#, longHelp="Zoom out")
        self.toolbar.AddSeparator()
        self.toolbar.AddTool(SIMFW_ID_ZOOM_MODE, "Zoom Mode", wx.Image("images/zoom.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Zoom Mode", kind=wx.ITEM_RADIO)
        self.toolbar.AddTool(SIMFW_ID_SELECT_MODE, "Select Mode", wx.Image("images/rectangle.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Select Mode", kind=wx.ITEM_RADIO)
        
        btnImg = wx.ArtProvider.GetBitmap(wx.ART_DELETE).ConvertToImage()
        btnImg.Rescale(20, 20)
        self.toolbar.AddTool(SIMFW_ID_DELETE, "Delete Selection", btnImg.ConvertToBitmap(), shortHelp="Delete selected tiles")
        self.toolbar.EnableTool(SIMFW_ID_DELETE, False)
        wx.ArtProvider.GetBitmap(wx.ART_DELETE)
        self.toolbar.AddTool(SIMFW_ID_UNDO, "Undo delete", wx.Image("images/undo.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Undo Delete")
        self.toolbar.EnableTool(SIMFW_ID_UNDO, False)
        self.toolbar.Realize()
        self.toolbar.Bind(wx.EVT_TOOL, self.OnTool)
        self.frame._mgr.AddPane(self.toolbar, wx.aui.AuiPaneInfo().
                          Name("Simulation toolbar").Caption("Simulation Toolbar").
                          ToolbarPane().Top())

    def OnTool(self, evt):
        if evt.GetId() == SIMFW_ID_STOP:
            self.canvas.DoQuit()
        elif evt.GetId() == SIMFW_ID_FOR:
            self.canvas.DoForwardStep()
        elif evt.GetId() == SIMFW_ID_FF:
            self.canvas.DoGo()
        elif evt.GetId() == SIMFW_ID_ZOOMIN:
            self.canvas.ZoomThenRecenter(True)
        elif evt.GetId() == SIMFW_ID_ZOOMOUT:
            self.canvas.ZoomThenRecenter(False)
        elif evt.GetId() == SIMFW_ID_ZOOM_MODE:
            self.SetMouseMode(SimCanvas.SIMFW_MODE_ZOOM)
        elif evt.GetId() == SIMFW_ID_SELECT_MODE:
            self.SetMouseMode(SimCanvas.SIMFW_MODE_SELECT)
        elif evt.GetId() == SIMFW_ID_DELETE:
            self.canvas.DeleteSelection()
        elif evt.GetId() == SIMFW_ID_UNDO:
            self.UndoDelete()
            
    def SetMouseMode(self, mode):
        self.canvas.mouseMode = mode
        if mode == SimCanvas.SIMFW_MODE_ZOOM:
            self.canvas.rubberBand.reset()
            self.canvas.SetCursor(wx.NullCursor)
            self.toolbar.EnableTool(SIMFW_ID_DELETE, False)
            self.toolbar.EnableTool(SIMFW_ID_UNDO, False)
        else:
            self.toolbar.EnableTool(SIMFW_ID_DELETE, True)
            self.toolbar.EnableTool(SIMFW_ID_UNDO, True)
        
    def UndoDelete(self):
        for loc in self.canvas.lastDeletedList:
            self.canvas.assembly.frontier.discard(loc)
            self.canvas.assembly.AddTile(loc.type, loc.coords)
        self.canvas.lastDeletedList = []
        self.canvas.DoRedrawAssembly()

    def OutputMessage(self, msg):
        self.messageQueue.append(msg)
        if (self.messageWin is None) or (not self.showMessages):
            return
        wx.CallAfter(pub.getDefaultPublisher().sendMessage, "outputMessage")
        
    def DoOutputMessage(self):
        if not self.messageWin is None:
            if len(self.messageQueue) == 0:
                return
            allMsgs = ""
            for msg in self.messageQueue:
                if len(msg) > 0:
                    if not msg.endswith("\n"):
                        msg = msg + "\n"
                    allMsgs = allMsgs + msg
            self.messageQueue = []

            self.messageWin.VisibleWrite(allMsgs)
            #self.messageWin.ScrollIntoView(self.messageWin.GetLastPosition())

    def OnExitApp(self, evt):
        dlg = wx.MessageDialog(self.canvas, "Are you sure you want to exit?")
        if dlg.ShowModal() == wx.ID_OK:
            dlg.Destroy()
            self.frame.Close(True)
        dlg.Destroy()
        
    def OnNew(self, evt):
        self.canvas.DoNew()

    def OnLoad(self, evt):
        self.canvas.DoLoad()
        
    def OnReload(self, evt):
        self.canvas.Reset(True)
        
    def OnReset(self, evt):
        self.canvas.Reset(False)

    def OnSave(self, event):
        self.canvas.DoSave()
        
    def OnForward(self, event):
        self.canvas.DoForwardStep()
        
    def OnGo(self, event):
        self.canvas.DoGo()
        
    def OnQuit(self, event):
        self.canvas.DoQuit()
            
    def OnToggleGrid(self, event):
        self.canvas.ToggleGrid()
        
    def OnToggleDisplay(self, event):
        self.canvas.ToggleDisplay()
        
    def OnShowMatching(self, event):
        self.canvas.SetShowMatching()
        
    def OnRefresh(self, event):
        self.canvas.Refresh()
             
    def OnZoomIn(self, event):
        self.canvas.ZoomThenRecenter(True)

    def OnZoomOut(self, event):
        self.canvas.ZoomThenRecenter(False)
        
    def OnCenter(self, event):
        self.canvas.ScrollToCoords((0,0))

    def OnHelp(self, event):
        self.canvas.ShowHelp()
        
    def ShowWarningDialog(self, msg):
        wx.CallAfter(pub.getDefaultPublisher().sendMessage, "warningDialog", msg=msg)
        
    def DoWarningDialog(self, msg):
        dlg = wx.MessageDialog(self.frame, msg, caption="Warning", style=wx.ICON_WARNING|wx.OK)
        dlg.ShowModal()
        dlg.Destroy()
        
    def ShowErrorDialog(self, msg):
        wx.CallAfter(pub.getDefaultPublisher().sendMessage, "errorDialog", msg=msg)

    def DoErrorDialog(self, msg):
        dlg = wx.MessageDialog(self.frame, msg, caption="Error", style=wx.ICON_ERROR|wx.OK)
        dlg.ShowModal()
        dlg.Destroy()

    def OnCloseFrame(self, evt):
        # Make sure the simulation thread is killed
        self.canvas.DoQuit()
        timer = wx.Timer(self)
        timer.Start(1000)
            #self.canvas.goThread.exit()

        # deinitialize the frame manager
        self.frame._mgr.UnInit()
        # delete the frame
        self.frame.Destroy()

#----------------------------------------------------------------------------
# The following lines create the actual application and begin its execution
# Include them (uncommented!) in your child class of SimApp
#app = SimApp()
#app.MainLoop()

#---------------------------------------------------------------------------


